function reviewedParams = rainSTORM_undrift_denoised(reviewedSupResParams,settings)
    % 0. Load required information
    markerAnchor = settings.undrift.markerAnchor;
    markerFrames = settings.undrift.markerFrames; % List of frame #s with mark
    markerPosits   = settings.undrift.markerPosits;
    reviewedParams = reviewedSupResParams;

    markerPosits = markerPosits - repmat(markerAnchor,size(markerPosits,1),1);

    nFrames = max( [reviewedParams.frame_idx] );
    
    drift = estimateDrift( nFrames, markerFrames, markerPosits );
    
    for lpLc = 1:length(reviewedParams)
        frameIx = reviewedParams(lpLc).frame_idx;
        
        reviewedParams(lpLc).x_coord = reviewedParams(lpLc).x_coord - drift(frameIx, 1);
        reviewedParams(lpLc).y_coord = reviewedParams(lpLc).y_coord - drift(frameIx, 2);
    end
end

%% Collects the drift information for each frame, and 
% removes the effect of the inaccuracy of the localization of the bead.
function [drift] = estimateDrift( nFrames, markerFrames, markerPosits )

    %% Get drift for each frame
    drift = zeros( nFrames, 2 );
    firstIx = find(markerFrames == min(markerFrames), 1, 'first' );
    lastPos = markerPosits( firstIx, : );
    for i = 1:nFrames
        ix = find( i==markerFrames, 1, 'first' );
        if numel( ix ) > 0
            lastPos = markerPosits( ix, : );
        end
        drift(i, :) = lastPos;
    end
    
    % Denoising drift vectors (by coordinates) to remove the inaccuracy
    % of the localization of the bead.
    drift(:,1) = denoise( drift(:,1) );
    drift(:,2) = denoise( drift(:,2) );
end

%% Denoising a vector to remove white noise.
% Convolution with different gaussian windows is performed.
% The residual corresponding to the minimal NCP will resemble the most
% to a white noise.
function y = denoise( x )
    N = length(x);
    
    % Considering Gaussian windows of different sizes
    % The smoothing vector with the minimal NCP value provides the result.
    sigmaList = 2.^[2:1:11];
    for i = 1:length(sigmaList)
        L = sigmaList(i);

        % Building Gaussian window
        ls = linspace(-(L-1)/L, (L-1)/L, L)';
        alpha = 2.5;
        G = exp( -(alpha*ls).^2/2 );
        G = G / sum(G);
    
        padded = padarray( x, [L/2 0], 'replicate' );
        
        y{i} = conv( padded, G, 'same' );
        y{i} = y{i}(L/2+1:end-L/2);
        
        r(i) = NCP(y{i}-x);                     % Computing the NCP of the residual and save it into 'r'
    end
    y = y{ find(r==min(r),1,'first') };         % The minimal NCP indicates the best smoothing.
end

%% NCP - Normalized Cumulative Periodogram
% Provides a measure of how a signal resemble to white-noise (the minimal
% NCP value indicates the more 'similarity' to a white-noise)
function R = NCP( f )
    n = length(f);
    p = ( abs(fft( f ) ) ).^2;
  
    OFFSET = 0;
    N = floor(n/2-1)-OFFSET;
    c = zeros(N,1);
    
    for k = 1:N
        c(k) = sum( abs( p( 2+OFFSET:k+OFFSET+1) ) ) / sum( abs( p(2+OFFSET:N+OFFSET+1) ) );
    end
    
    line = (1:N)' / (N) * max(c);
    
    R = sum(abs(c-line).^2);
end